"""Helpers to format/display various information."""

import click


def display_changeset(changeset):
    for change in changeset['Changes']:
        if change['Type'] != 'Resource':
            continue
        action = change['ResourceChange']['Action']
        resource_type = change['ResourceChange']['ResourceType']
        resource_id = change['ResourceChange']['LogicalResourceId']
        click.echo(f"{resource_type} {resource_id} -> {action}")



"""Helpers to interact with the config file."""
import os

import yaml


class Config:
    def __init__(self, root, raw_config):
        self.root = root
        self.raw_config = raw_config

    def get_stack(self, name):
        stack = self.raw_config.get(name)
        if stack is None:
            raise Exception(f"Could not find stack config named '{name}'")

        stack_name = stack.get('stack')
        stack_file = stack.get('file')
        use_changesets = stack.get('use_changesets')

        if not os.path.isabs(stack_file):
            stack_file = os.path.join(self.root, stack_file)

        if not os.path.exists(stack_file):
            raise Exception(f"Could not find stack file: '{stack_file}'")

        return stack_name, stack_file, use_changesets


def read_config(config_file_name):
    config_file = os.path.abspath(config_file_name)
    root = os.path.dirname(config_file)

    try:
        with open(config_file, 'r') as stream:
            config = yaml.safe_load(stream)
            return Config(root, config)
    except Exception:
        return None



"""Command used to deploy/update a cloudformation stack."""

import click


from cloudformation_helper.utils.aws import (
    create_changeset,
    delete_changeset,
    execute_changeset,
    get_changeset,
    has_changeset,
    stack_exists,
    create_stack,
    update_stack,
)
from cloudformation_helper.utils.formatter import (
    display_changeset,
)


def update_using_changeset(stack_name, stack_file):
    click.echo('Creating changeset for existing stack')
    create_changeset(stack_name, stack_file, False)
    changeset = get_changeset(stack_name)
    display_changeset(changeset)
    if click.confirm('Execute stack changes? ', default=False):
        execute_changeset(stack_name, False)
    else:
        if click.confirm('Keep pending changes? ', default=False):
            click.echo('Aborted')
        else:
            delete_changeset(stack_name)


def create_using_changeset(stack_name, stack_file):
    click.echo('Creating changeset for new stack')
    create_changeset(stack_name, stack_file, True)
    changeset = get_changeset(stack_name)
    display_changeset(changeset)
    if click.confirm('Execute stack changes? ', default=False):
        execute_changeset(stack_name, True)
    else:
        if click.confirm('Keep pending changes? ', default=False):
            click.echo('Aborted')
        else:
            delete_changeset(stack_name)


def deploy_or_update(stack_name, stack_file, use_changesets):
    click.echo(f"Processing {stack_name} using {stack_file}")

    if use_changesets and has_changeset(stack_name):
        if click.confirm('Stack already has a changeset; delete it? ', default=False):
            click.echo(f"Deleting changeset for stack {stack_name}")
            delete_changeset(stack_name)
        else:
            click.echo('Aborted')
            return

    if stack_exists(stack_name):
        if use_changesets:
            update_using_changeset(stack_name, stack_file)
        else:
            update_stack(stack_name, stack_file)
    else:
        if use_changesets:
            create_using_changeset(stack_name, stack_file)
        else:
            create_stack(stack_name, stack_file)



"""Helpers to access AWS information."""

import botocore
import boto3


CHANGESET_NAME = 'cfhelper-changeset'


def stack_exists(stack_name):
    client = boto3.client('cloudformation')

    try:
        client.describe_stacks(StackName=stack_name)
    except botocore.exceptions.ClientError as error:
        if error.response['Error']['Code'] == 'ValidationError' and 'does not exist' in error.response['Error']['Message']:
            return False
        raise error

    return True


def create_stack(stack_name, stack_file):
    client = boto3.client('cloudformation')
    with open(stack_file) as f:
        template = f.read()

    client.create_stack(
        StackName=stack_name,
        TemplateBody=template,
        Capabilities=[
            'CAPABILITY_IAM',
        ],
    )
    waiter = client.get_waiter('stack_create_complete')

    waiter.wait(
        StackName=stack_name,
        WaiterConfig={
            'Delay': 15,
            'MaxAttempts': 120
        }
    )


def update_stack(stack_name, stack_file):
    client = boto3.client('cloudformation')
    with open(stack_file) as f:
        template = f.read()

    # TODO, handle errors
    # botocore.errorfactory.InsufficientCapabilitiesException: An error occurred (InsufficientCapabilitiesException)
    # when calling the UpdateStack operation: Requires capabilities : [CAPABILITY_IAM]
    # botocore.exceptions.ClientError: An error occurred (ValidationError) when calling the UpdateStack operation: No updates are to be performed.
    client.update_stack(
        StackName=stack_name,
        TemplateBody=template,
        Capabilities=[
            'CAPABILITY_IAM',
        ],
    )

    waiter = client.get_waiter('stack_update_complete')
    waiter.wait(
        StackName=stack_name,
        WaiterConfig={
            'Delay': 15,
            'MaxAttempts': 120
        }
    )


def has_changeset(stack_name):
    try:
        get_changeset(stack_name)
    except botocore.exceptions.ClientError as error:
        if error.response['Error']['Code'] == 'ValidationError' and 'does not exist' in error.response['Error']['Message']:
            return False
        if error.response['Error']['Code'] == 'ChangeSetNotFound':
            return False
        raise error

    return True


def get_changeset(stack_name):
    client = boto3.client('cloudformation')

    return client.describe_change_set(
        StackName=stack_name,
        ChangeSetName=CHANGESET_NAME,
    )


def create_changeset(stack_name, stack_file, is_creation):
    client = boto3.client('cloudformation')
    changeset_type = 'CREATE' if is_creation else 'UPDATE'
    with open(stack_file) as f:
        template = f.read()

    client.create_change_set(
        StackName=stack_name,
        TemplateBody=template,
        Capabilities=[
            'CAPABILITY_IAM',
        ],
        ChangeSetName=CHANGESET_NAME,
        ChangeSetType=changeset_type,
    )

    waiter = client.get_waiter('change_set_create_complete')
    waiter.wait(
        ChangeSetName=CHANGESET_NAME,
        StackName=stack_name,
        WaiterConfig={
            'Delay': 15,
            'MaxAttempts': 120
        }
    )


def execute_changeset(stack_name, is_creation):
    client = boto3.client('cloudformation')
    client.execute_change_set(
        StackName=stack_name,
        ChangeSetName=CHANGESET_NAME,
    )

    waiter_type = 'stack_create_complete' if is_creation else 'stack_update_complete'
    waiter = client.get_waiter(waiter_type)
    waiter.wait(
        StackName=stack_name,
        WaiterConfig={
            'Delay': 15,
            'MaxAttempts': 120
        }
    )


def delete_changeset(stack_name):
    client = boto3.client('cloudformation')

    client.delete_change_set(
        StackName=stack_name,
        ChangeSetName=CHANGESET_NAME,
    )



"""Console script for cloudformation_helper."""
import sys
import click

from cloudformation_helper.commands.deploy import deploy_or_update
from cloudformation_helper.utils.config import read_config


@click.group()
@click.option('--config', default='stacks.cfh')
@click.pass_context
def cfhelper(ctx, config):
    ctx.obj = read_config(config)


@cfhelper.command()
@click.argument("args", nargs=-1)
@click.pass_obj
def deploy(config, args):
    if len(args) == 1:
        if not config:
            raise Exception('Could not find configuration file')
        config_name, = args
        stack_name, stack_file, use_changesets = config.get_stack(config_name)
        deploy_or_update(stack_name, stack_file, use_changesets)
    elif len(args) == 2:
        stack_name, stack_file = args
        deploy_or_update(stack_name, stack_file, True)
    else:
        click.echo('Missing arguments')


def run():
    sys.exit(cfhelper(auto_envvar_prefix='CFHELPER'))  # pragma: no cover


if __name__ == "__main__":
    run()
